import json
from copy import deepcopy
from datetime import datetime
from io import BytesIO
from typing import Dict, List, Union

import pandas as pd

from cognite import CogniteClient
from cognite.client._utils import get_aggregate_func_return_name, to_camel_case, to_snake_case


class TimeSeries:
    """Object for specifying a specific time series from the TimeSeries API when using a data spec.

    Args:
        id (int): id of time series to retrieve
        aggregates (List[str]): Local aggregate functions to apply
        missing_data_strategy(str): Missing data strategy to apply
        label (str): name of the column in the resulting data frame when passed to data transfer service.

    Examples:
        When you supply a label the resulting data frames produced by data transfer service
        will use the label as column names.
    """

    def __init__(self, id: int, aggregates: List[str] = None, missing_data_strategy: str = None, label: str = None):
        self.id = id
        self.aggregates = aggregates
        self.missing_data_strategy = missing_data_strategy
        self.label = label
        self._name = None


class TimeSeriesDataSpec:
    """Object for specifying data from the TimeSeries API when using a data spec.

    Args:
        time_series (List[data_transfer_service.TimeSeries]):  Time series
        aggregates (List[str]): List of aggregate functions
        granularity (str): Granularity of aggregates
        missing_data_strategy (str): Missing data strategy to apply, can be "linearInterpolation" or "ffill"
        start (Union[str, int, datetime]): Start time
        end (Union[str, int, datetime]): end time
        label (str): Label for this data spec

    Examples:
        When you specify a label, data transfer service will exhibit the following behaviour::

            ts_data_spec = TimeSeriesDataSpec(..., label="some_label")
            data_spec = DataSpec([ts_data_spec], ...)
            dts = DataTransferService(data_spec)
            dataframes = dts.get_dataframes
            my_df = dataframes["some_label"]
    """

    def __init__(
        self,
        time_series: List[TimeSeries],
        aggregates: List[str],
        granularity: str,
        missing_data_strategy: str = None,
        start: Union[str, int, datetime] = None,
        end: Union[str, int, datetime] = None,
        label: str = None,
    ):
        self.time_series = time_series
        self.aggregates = aggregates
        self.granularity = granularity
        self.missing_data_strategy = missing_data_strategy
        self.start = start
        self.end = end
        self.label = label or "default"


class FilesDataSpec:
    """Object for specifying data from the Files API when using a data spec.

    Args:
        file_ids (Dict[str, int]):    Dictionary of fileNames -> fileIds
    """

    def __init__(self, file_ids: Dict[str, int]):
        self.file_ids = file_ids


class DataSpec:
    """Object for specifying data when querying CDP.

    Args:
        time_series_data_specs (List[TimeSeriesDataSpec]):  Time Series data specs
        files_data_spec (FilesDataSpec):                    Files data spec

    Raises:
        DataSpecValidationError: An error occurred while validating the data spec

    """

    def __init__(self, time_series_data_specs: List[TimeSeriesDataSpec] = None, files_data_spec: FilesDataSpec = None):
        self.time_series_data_specs = time_series_data_specs
        self.files_data_spec = files_data_spec
        self.__validate_time_series_data_specs()
        self.__validate_files_data_spec()

    def __validate_time_series_data_specs(self):
        if self.time_series_data_specs:
            labels = []
            for tsds in self.time_series_data_specs:
                if not isinstance(tsds, TimeSeriesDataSpec):
                    raise DataSpecValidationError("Time series data specs must be TimeSeriesDataSpec objects.")

                label = tsds.label or "default"
                if label in labels:
                    raise DataSpecValidationError("Unique labels for each dataspec must be used")
                labels.append(label)

                if not isinstance(tsds.time_series, list):
                    raise DataSpecValidationError("Time series must be a list of time series objects.")

                if not len(tsds.time_series) > 0:
                    raise DataSpecValidationError("A time series data spec does not contain any time series")

                ts_labels = []
                for ts in tsds.time_series:
                    if not isinstance(ts, TimeSeries):
                        raise DataSpecValidationError("Time series must be a TimeSeries object")
                    if ts.label and ts.label in ts_labels:
                        raise DataSpecValidationError("Time series labels must be unique")
                    ts_labels.append(ts.label)

    def __validate_files_data_spec(self):
        if self.files_data_spec:
            if not isinstance(self.files_data_spec, FilesDataSpec):
                raise DataSpecValidationError("files_data_spec must be a FilesDataSpec object")

            if not isinstance(self.files_data_spec.file_ids, dict):
                raise DataSpecValidationError("file_ids must be a dict of form {name: id}")

            for name, id in self.files_data_spec.file_ids.items():
                if not isinstance(name, str):
                    raise DataSpecValidationError("File names must be a strings")
                if not isinstance(id, int):
                    raise DataSpecValidationError("File ids must be integers")

    def to_JSON(self):
        return DataSpec._to_json(self)

    def __str__(self):
        return json.dumps(self.to_JSON(), indent=4)

    @staticmethod
    def _to_json(obj):
        if isinstance(obj, (DataSpec, TimeSeries, FilesDataSpec, TimeSeriesDataSpec)):
            return DataSpec._to_json(obj.__dict__)
        if isinstance(obj, dict):
            return {to_camel_case(key): DataSpec._to_json(value) for key, value in obj.items() if (value is not None)}
        elif isinstance(obj, list):
            new_list = []
            for el in obj:
                new_list.append(DataSpec._to_json(el))
            return new_list
        elif isinstance(obj, (str, int, float, bool)) or obj is None:
            return obj
        raise AssertionError("Data spec does not accept type {}".format(type(obj)))

    @classmethod
    def from_JSON(cls, json_repr):
        if isinstance(json_repr, str):
            json_repr = json.loads(json_repr)

        if not isinstance(json_repr, dict):
            raise DataSpecValidationError("from_JSON accepts a dict")

        time_series_data_specs_json = json_repr.get("timeSeriesDataSpecs")
        files_data_spec_json = json_repr.get("filesDataSpec")

        if not (time_series_data_specs_json or files_data_spec_json):
            raise DataSpecValidationError("Data spec must include at least one of [timeSeriesDataSpec, filesDataSpec]")

        time_series_data_specs = []
        if time_series_data_specs_json:
            for tsds_json in time_series_data_specs_json:
                time_series_json = tsds_json.get("timeSeries")
                if time_series_json:
                    tsds = TimeSeriesDataSpec(**{to_snake_case(key): value for key, value in tsds_json.items()})
                    tsds.time_series = [
                        TimeSeries(**{to_snake_case(key): value for key, value in ts.items()})
                        for ts in time_series_json
                    ]
                    time_series_data_specs.append(tsds)

        files_data_spec = None
        if files_data_spec_json:
            files_data_spec = FilesDataSpec(
                **{to_snake_case(key): value for key, value in files_data_spec_json.items()}
            )
        return DataSpec(time_series_data_specs, files_data_spec)


class DataSpecValidationError(Exception):
    pass


class DataTransferService:
    """Create a Data Transfer Service object.

    Fetch timeseries from the api.
    """

    def __init__(self, data_spec: DataSpec, api_key: str = None, cookies: Dict = None, num_of_workers: int = 10):
        """
        Args:
            data_spec (data_transfer_service.DataSpec):   Data Spec.
            api_key (str):          Api key.
            cookies (dict):         Cookies.
            num_of_workers (int):   Number of workers to fetch data with.
        """
        self.cognite_client = CogniteClient(api_key=api_key, cookies=cookies, num_of_workers=num_of_workers)

        if isinstance(data_spec, DataSpec):
            self.data_spec = deepcopy(data_spec)
        elif isinstance(data_spec, dict):
            self.data_spec = DataSpec.from_JSON(data_spec)
        else:
            raise ValueError("DataTransferService accepts a DataSpec instance or a json object representation of it.")
        self.ts_data_specs = self.data_spec.time_series_data_specs
        self.files_data_spec = self.data_spec.files_data_spec
        self.cookies = cookies

    def get_time_series_name(self, ts_label: str, dataframe_label: str = "default"):
        if self.ts_data_specs is None:
            raise ValueError("Data spec does not contain any TimeSeriesDataSpecs")

        tsds = None
        for ts_data_spec in self.ts_data_specs:
            if ts_data_spec.label == dataframe_label:
                tsds = ts_data_spec

        if tsds:
            # Temporary workaround that you cannot use get_datapoints_frame with ts id.
            ts_res = self.cognite_client.experimental.time_series.get_multiple_time_series_by_id(
                ids=list(set([ts.id for ts in tsds.time_series]))
            )
            id_to_name = {ts["id"]: ts["name"] for ts in ts_res.to_json()}

            for ts in tsds.time_series:
                if ts.label == ts_label:
                    return id_to_name[ts.id]
            raise ValueError("Invalid time series label")
        raise ValueError("Invalid dataframe label")

    def get_dataframes(self, drop_agg_suffix: bool = True):
        """Return a dictionary of dataframes indexed by label - one per data spec.

        Args:
            drop_agg_suffix (bool): If a time series has only one aggregate, drop the `|<agg-func>` suffix on
                                    those column names.
        Returns:
            Dict[str, pd.DataFrame]: A label-indexed dictionary of data frames.
        """
        if len(self.ts_data_specs) == 0 or self.ts_data_specs is None:
            raise ValueError("Data spec does not contain any TimeSeriesDataSpecs")

        dataframes = {}
        for tsds in self.ts_data_specs:
            df = self.get_dataframe(tsds.label, drop_agg_suffix=drop_agg_suffix)
            dataframes[tsds.label] = df
        return dataframes

    def get_dataframe(self, label: str = "default", drop_agg_suffix: bool = True):
        if self.ts_data_specs is None:
            raise ValueError("Data spec does not contain any TimeSeriesDataSpecs")

        tsds = None
        for ts_data_spec in self.ts_data_specs:
            if ts_data_spec.label == label:
                tsds = ts_data_spec
        if tsds:
            ts_list = []
            # Temporary workaround that you cannot use get_datapoints_frame with ts id.
            ts_res = self.cognite_client.experimental.time_series.get_multiple_time_series_by_id(
                ids=list(set([ts.id for ts in tsds.time_series]))
            )
            id_to_name = {ts["id"]: ts["name"] for ts in ts_res.to_json()}

            for ts in tsds.time_series:
                ts._name = id_to_name[ts.id]
                if not ts.label:
                    ts.label = id_to_name[ts.id]
                ts_dict = dict(
                    name=id_to_name[ts.id],
                    aggregates=ts.aggregates or tsds.aggregates,
                    missingDataStrategy=ts.missing_data_strategy,
                )
                ts_list.append(ts_dict)

            df = self.cognite_client.datapoints.get_datapoints_frame(
                ts_list, tsds.aggregates, tsds.granularity, tsds.start, tsds.end
            )
            df = self.__apply_missing_data_strategies(df, ts_list, tsds.missing_data_strategy)
            df = self.__convert_ts_names_to_labels(df, tsds, drop_agg_suffix)
            return df
        raise ValueError("Invalid label")

    def get_file(self, name):
        """Return files by name as specified in the DataSpec

        Args:
            name (str): Name of file
        """
        if not self.files_data_spec or not isinstance(self.files_data_spec, FilesDataSpec):
            raise ValueError("Data spec does not contain a FilesDataSpec")
        id = self.files_data_spec.file_ids.get(name)
        if id:
            file_bytes = self.cognite_client.files.download_file(id, get_contents=True)
            return BytesIO(file_bytes)
        raise ValueError("Invalid name")

    def __convert_ts_names_to_labels(self, df: pd.DataFrame, tsds: TimeSeriesDataSpec, drop_agg_suffix: bool):
        name_to_label = {}
        for ts in tsds.time_series:
            for agg in ts.aggregates or tsds.aggregates:
                agg_return_name = get_aggregate_func_return_name(agg)
                if len(ts.aggregates or tsds.aggregates) > 1 or not drop_agg_suffix:
                    name_to_label[ts._name + "|" + agg_return_name] = ts.label + "|" + agg_return_name
                else:
                    name_to_label[ts._name + "|" + agg_return_name] = ts.label
        return df.rename(columns=name_to_label)

    def __apply_missing_data_strategies(self, df, ts_list, global_missing_data_strategy):
        """Applies missing data strategies to dataframe.

        Local strategies have precedence over global strategy.
        """
        new_df = df["timestamp"]
        for ts in ts_list:
            name = ts["name"]
            colnames = [colname for colname in df.columns.values if colname.startswith(name)]

            missing_data_strategy = ts.get("missingDataStrategy", global_missing_data_strategy)
            partial_df = df[colnames]

            if missing_data_strategy == "ffill":
                partial_df = df[colnames].fillna(method="pad")
            elif missing_data_strategy and missing_data_strategy.endswith("Interpolation"):
                method = missing_data_strategy[:-13].lower()
                partial_df = df[colnames].interpolate(method=method, axis=0)
            new_df = pd.concat([new_df, partial_df], axis=1)
            new_df = new_df.iloc[:, ~new_df.columns.duplicated()]
        return new_df



from cognite.client.experimental.analytics import AnalyticsClient
from cognite.client.experimental.datapoints import DatapointsClient
from cognite.client.experimental.sequences import SequencesClient
from cognite.client.experimental.time_series import TimeSeriesClient


class ExperimentalClient:
    def __init__(self, client_factory):
        self._client_factory = client_factory

    @property
    def analytics(self) -> AnalyticsClient:
        return AnalyticsClient(client_factory=self._client_factory)

    @property
    def datapoints(self) -> DatapointsClient:
        return self._client_factory(DatapointsClient)

    @property
    def sequences(self) -> SequencesClient:
        return self._client_factory(SequencesClient)

    @property
    def time_series(self) -> TimeSeriesClient:
        return self._client_factory(TimeSeriesClient)



# -*- coding: utf-8 -*-
import json
from typing import Dict, List

import pandas as pd

from cognite.client._api_client import APIClient, CogniteResponse


class AssetListResponse(CogniteResponse):
    """Assets Response Object"""

    def __init__(self, internal_representation):
        super().__init__(internal_representation)
        self.counter = 0

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        if len(self.to_json()) > 0:
            return pd.DataFrame(self.internal_representation["data"]["items"])
        return pd.DataFrame()

    def __iter__(self):
        return self

    def __next__(self):
        if self.counter > len(self.to_json()) - 1:
            raise StopIteration
        else:
            self.counter += 1
            return AssetResponse({"data": {"items": [self.to_json()[self.counter - 1]]}})


class AssetResponse(CogniteResponse):
    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation["data"]["items"][0]

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        if len(self.to_json()) > 0:
            return pd.DataFrame.from_dict(self.to_json(), orient="index")
        return pd.DataFrame()


class Asset:
    """Data transfer object for assets.

    Args:
        name (str):                 Name of asset. Often referred to as tag.
        parent_id (int):            ID of parent asset, if any.
        description (str):          Description of asset.
        metadata (dict):            Custom , application specific metadata. String key -> String Value.
        ref_id (str):               Reference ID used only in post request to disambiguate references to duplicate
                                    names.
        parent_name (str):          Name of parent, this parent must exist in the same POST request.
        parent_ref_id (str):        Reference ID of parent, to disambiguate if multiple nodes have the same name.
    """

    def __init__(
        self, name, parent_id=None, description=None, metadata=None, ref_id=None, parent_name=None, parent_ref_id=None
    ):
        self.name = name
        self.parentId = parent_id
        self.description = description
        self.metadata = metadata
        self.refId = ref_id
        self.parentName = parent_name
        self.parentRefId = parent_ref_id


class AssetsClient(APIClient):
    def __init__(self, **kwargs):
        super().__init__(version="0.5", **kwargs)

    def get_assets(
        self, name=None, path=None, description=None, metadata=None, depth=None, fuzziness=None, **kwargs
    ) -> AssetListResponse:
        """Returns assets matching provided description.

        Args:
            name (str):             The name of the asset(s) to get.

            path (str):             The path of the subtree to search in.

            description (str):      Search query.

            metadata (dict):         The metadata values used to filter the results.

            depth (int):            Get sub assets up oto this many levels below the specified path.

            fuzziness (int):        The degree of fuzziness in the name matching.

        Keyword Arguments:
            autopaging (bool):      Whether or not to automatically page through results. If set to true, limit will be
                                    disregarded. Defaults to False.

            limit (int):            The maximum number of assets to be returned.

            cursor (str):           Cursor to use for paging through results.
        Returns:
            stable.assets.AssetListResponse: A data object containing the requested assets with several getter methods with different
            output formats.

        Examples:
            You can fetch all assets with a maximum depth of three like this::

                client = CogniteClient()
                res = client.assets.get_assets(depth=3, autopaging=True)
                print(res.to_pandas())
        """
        url = "/assets"
        params = {
            "name": name,
            "description": description,
            "path": path,
            "metadata": str(metadata) if metadata else None,
            "depth": depth,
            "fuzziness": fuzziness,
            "cursor": kwargs.get("cursor"),
            "limit": kwargs.get("limit", self._LIMIT) if not kwargs.get("autopaging") else self._LIMIT,
        }
        res = self._get(url, params=params)
        assets = []
        assets.extend(res.json()["data"]["items"])
        next_cursor = res.json()["data"].get("nextCursor")

        while next_cursor and kwargs.get("autopaging"):
            params["cursor"] = next_cursor
            res = self._get(url=url, params=params)
            assets.extend(res.json()["data"]["items"])
            next_cursor = res.json()["data"].get("nextCursor")

        return AssetListResponse(
            {
                "data": {
                    "nextCursor": next_cursor,
                    "previousCursor": res.json()["data"].get("previousCursor"),
                    "items": assets,
                }
            }
        )

    def get_asset(self, asset_id) -> AssetResponse:
        """Returns the asset with the provided assetId.

        Args:
            asset_id (int):         The asset id of the top asset to get.

        Returns:
            stable.assets.AssetResponse: A data object containing the requested assets with several getter methods with different
            output formats.
        Examples:
            You can fetch a single asset like this::

                client = CogniteClient()
                res = client.assets.get_asset(asset_id=123)
                print(res)
        """
        url = "/assets/{}/subtree".format(asset_id)
        res = self._get(url)
        return AssetResponse(res.json())

    def get_asset_subtree(self, asset_id, depth=None, **kwargs) -> AssetListResponse:
        """Returns asset subtree of asset with provided assetId.

        Args:
            asset_id (int):         The asset id of the top asset to get.

            depth (int):            Get subassets this many levels below the top asset.

        Keyword Arguments:
            limit (int):            The maximum nuber of assets to be returned.

            cursor (str):           Cursor to use for paging through results.
        Returns:
            stable.assets.AssetListResponse: A data object containing the requested assets with several getter methods with different
            output formats.

        Examples:
            You can fetch an asset subtree like this::

                client = CogniteClient()
                res = client.assets.get_asset_subtree(asset_id=123, depth=)
                print(res.to_pandas())
        """
        url = "/assets/{}/subtree".format(asset_id)
        params = {"depth": depth, "limit": kwargs.get("limit", self._LIMIT), "cursor": kwargs.get("cursor")}
        res = self._get(url, params=params)
        return AssetListResponse(res.json())

    def post_assets(self, assets: List[Asset]) -> AssetListResponse:
        """Insert a list of assets.

        Args:
            assets (list[stable.assets.Asset]): List of asset data transfer objects.

        Returns:
            stable.assets.AssetListResponse: A data object containing the posted assets with several getter methods with different
            output formats.

        Examples:
            Posting an asset::

                from cognite.client.stable.assets import Asset

                client = CogniteClient()

                my_asset = Asset("myasset")
                assets_to_post = [my_asset]
                res = client.assets.post_assets(assets_to_post)
                print(res)
        """
        url = "/assets"
        body = {"items": [asset.__dict__ for asset in assets]}
        res = self._post(url, body=body)
        return AssetListResponse(res.json())

    def delete_assets(self, asset_ids: List[int]) -> None:
        """Delete a list of assets.

        Args:
            asset_ids (list[int]): List of IDs of assets to delete.

        Returns:
            None

        Examples:
            You can delete an asset like this::

                client = CogniteClient()
                res = client.assets.delete_assets([123])
        """
        url = "/assets/delete"
        body = {"items": asset_ids}
        self._post(url, body=body)

    def search_for_assets(
        self,
        name=None,
        description=None,
        query=None,
        metadata=None,
        asset_subtrees=None,
        min_created_time=None,
        max_created_time=None,
        min_last_updated_time=None,
        max_last_updated_time=None,
        **kwargs
    ) -> AssetListResponse:
        """Search for assets.

        Args:
            name:   Prefix and fuzzy search on name.
            description str:   Prefix and fuzzy search on description.
            query (str):       Search on name and description using wildcard search on each of the words
                                (separated by spaces). Retrieves results where at least one word must match.
                                Example: 'some other'
            metadata (dict):        Filter out assets that do not match these metadata fields and values (case-sensitive).
                                    Format is {"key1":"value1","key2":"value2"}.
            asset_subtrees (List[int]): Filter out assets that are not linked to assets in the subtree rooted at these assets.
                                        Format is [12,345,6,7890].
            min_created_time(str):  Filter out assets with createdTime before this. Format is milliseconds since epoch.
            max_created_time (str): Filter out assets with createdTime after this. Format is milliseconds since epoch.
            min_last_updated_time(str):  Filter out assets with lastUpdatedtime before this. Format is milliseconds since epoch.
            max_last_updated_time(str): Filter out assets with lastUpdatedtime after this. Format is milliseconds since epoch.

        Keyword Args:
            sort (str):             Field to be sorted.
            dir (str):              Sort direction (desc or asc)
            limit (int):            Return up to this many results. Max is 1000, default is 25.
            offset (int):           Offset from the first result. Sum of limit and offset must not exceed 1000. Default is 0.
            boost_name (str):       Whether or not boosting name field. This option is test_experimental and can be changed.
        Returns:
            stable.assets.AssetListResponse.

        Examples:
            Searching for assets::

                client = CogniteClient()
                res = client.assets.search_for_assets(name="myasset")
                print(res)
        """
        url = "/assets/search"
        params = {
            "name": name,
            "description": description,
            "query": query,
            "metadata": json.dumps(metadata),
            "assetSubtrees": asset_subtrees,
            "minCreatedTime": min_created_time,
            "maxCreatedTime": max_created_time,
            "minLastUpdatedTime": min_last_updated_time,
            "maxLastUpdatedTime": max_last_updated_time,
            "sort": kwargs.get("sort"),
            "dir": kwargs.get("dir"),
            "limit": kwargs.get("limit", 1000),
            "offset": kwargs.get("offset"),
            "boostName": kwargs.get("boost_name"),
        }

        res = self._get(url, params=params)
        return AssetListResponse(res.json())
